xquery version "3.1";

(:~
 : This module is responsible for retrieving the variants of a given text on a
 : given page. For this, it leverages the tokenization of the text that takes
 : place during commons:get-page-fragment.
 : 
 : The output of the main function, vars:get-variants, is a sequence of maps of
 : the following format:
 :
 : map {
 :  "body" 
 :  map {
 :      "x-content-type" : "Variant",
 :      "value" : [ map {
 :          "entry" : "${text-of-respective-witness}",
 :          "witness" : "${ms-id}"
 :      }],
 :      "format" : "text/plain",
 :      "type" : "TextualBody"
 :    },
 :    "target" : map {
 :      "format" : "text/xml",
 :      "language" : "syc",
 :      "id" : "http://ahikar.sub.uni-goettingen.de/ns/annotations/${teixml-uri}/${token-id}"
 :    },
 :    "type" : "Annotation",
 :    "id" : "http://ahikar.sub.uni-goettingen.de/ns/annotations/${teixml-uri}/annotation-variants-${token-id}"
 : }
 : 
 :)

module namespace vars="http://ahiqar.uni-goettingen.de/ns/annotations/variants";

declare namespace tei="http://www.tei-c.org/ns/1.0";

import module namespace commons="http://ahiqar.uni-goettingen.de/ns/commons" at "../commons.xqm";

declare variable $vars:ns := "http://ahiqar.uni-goettingen.de/ns/annotations";

(:~

:)
declare function vars:get-variants($pages as element(tei:TEI)+,
    $teixml-uri as xs:string,
    $pb-n as xs:string)
as map(*)* {
    for $page in $pages return
        let $variants := vars:get-variants-on-page-as-maps($teixml-uri, $page, $pb-n)
        for $map in $variants
            for $variant at $pos in $map?("variants")
            let $tokenId := (map:get($map, "current") => map:get("id"))
            let $id := $tokenId || "-" || $pos
            return
                map {
                    "id": $vars:ns || "/" || $teixml-uri || "/annotation-variants-" || $id,
                    "type": "Annotation",
                    "body": vars:get-body-object($variant),
                    "target": array{ vars:get-target-information($teixml-uri, $tokenId, $page) }
                }
};

(:~
 : Returns a map holding the current token as well as all its resp. variants per
 : page. The exact value of the current token depends on the manuscript we are
 : currently looking at.
 : 
 : @param $teixml-uri The current TEI document's base URI, e.g. "12345"
 : @param $page The page as given in tei:pb/@n
 : @return A map with information about the current token and its variants 
 :)
declare function vars:get-variants-on-page-as-maps($teixml-uri as xs:string,
    $page as element(tei:TEI),
    $pb-n as xs:string)
as map()* {
    let $ms-id := vars:get-ms-id-from-idno($teixml-uri)
    let $relevant-files-for-ms-id := vars:get-relevant-files($ms-id,$teixml-uri, $pb-n)
    let $tokens := vars:get-token-ids-on-page($teixml-uri, $pb-n)
    let $files-relevant-for-page := vars:get-files-relevant-for-page($relevant-files-for-ms-id, $tokens)
    
    return if (count($files-relevant-for-page) gt 0) then
        let $ms-id-position := vars:determine-id-position($ms-id, $files-relevant-for-page[1])
        let $non-ms-id-positions := vars:get-non-ms-id-positions-in-array($files-relevant-for-page[1], $ms-id-position)
        for $file in $files-relevant-for-page return
            let $indices-relevant-for-page := vars:get-indices-relevant-for-page($file, $ms-id-position, $tokens)
            
            for $iii in $indices-relevant-for-page return
                vars:make-map-for-token($file, $iii, $ms-id-position, $non-ms-id-positions)
    else
        util:log-system-out("WARN: No relevant files found for this page. This might be an error.")
};

declare function vars:get-ms-id-from-idno($teixml-uri as xs:string)
as xs:string {
    let $TEI := commons:open-tei-xml($teixml-uri)//tei:TEI
    return
        if (matches($TEI//tei:monogr/tei:editor, "Salhani")) then
            "Salhani"
        else
            commons:make-id-from-idno($TEI)
            => substring-after("t_")
};

declare function vars:get-relevant-files($ms-id as xs:string,
    $teixml-uri as xs:string,
    $pb-n as xs:string)
as item()+ {
    let $milestones-for-page := local:get-milestones-for-page($teixml-uri, $pb-n)
    let $collection-uri := "/db/data/collation-results/"
    let $collation-uris := xmldb:get-child-resources($collection-uri)
    let $relevant-uris :=
        for $m in $milestones-for-page return
            let $type := $m/@unit
            for $uri in $collation-uris return
                if(contains($uri, replace($ms-id, "/", ""))
                and contains($uri, $type)
                and ends-with($uri, "xml")) then
                    $uri
                else
                    ()
    for $uri in $relevant-uris return
        doc($collection-uri || $uri)
};

declare function local:get-milestones-for-page($teixml-uri as xs:string,
    $pb-n as xs:string) {
    let $group := doc($commons:data || $teixml-uri || ".xml")/tei:TEI/tei:text/tei:group
    let $relevant-text-type :=
        if ($group/tei:text[@xml:lang = 'ara']) then
            $group/tei:text[@xml:lang = 'ara']/@type
        else
            'transcription'
    let $pb := $group/tei:text[@type = $relevant-text-type]/tei:body//tei:pb[@n = $pb-n]
    let $next-pb := 
        if ($pb/following::tei:pb[1][ancestor::tei:text[$relevant-text-type]]) then
            $pb/following::tei:pb[1][ancestor::tei:text[$relevant-text-type]]
        else
            $pb/ancestor::tei:text[@type = $relevant-text-type]/tei:body/descendant::element()[last()]
    
    let $prev-milestone :=
        if (not($pb/following::element()[ancestor::tei:text[@type = $relevant-text-type]][. << $next-pb][self::tei:milestone])
            or not($pb/following::element()[1][self::tei:milestone])) then
            $pb/preceding::tei:milestone[1][ancestor::tei:text[@type = $relevant-text-type]]
        else
            ()
    let $milestones-on-page := $pb/following::element()[ancestor::tei:text[@type = $relevant-text-type]][. << $next-pb][self::tei:milestone]
    
    return ($prev-milestone, $milestones-on-page)
};

(:~
 : In the CollateX results, we first have an aray of witnesses before the result
 : of the collation is given in the JSON file. The position of a MS identifier
 : in this array is reflected in the collation result; The witness on position 2
 : will always have its tokens in the second position of the result.
 : 
 : @param $ms-id The MS identifier as given in the array of witnesses
 : @param $json The collation result as JSON
 : @return The position of the current MS identifier in the witnesses array
 :)
declare function vars:determine-id-position($ms-id as xs:string,
    $file as item())
as xs:integer {
    let $t_ms-id := "t_" || replace($ms-id, "[/]", "_")
    let $witness := $file//fn:array[@key = "witnesses"]/fn:string[text() = $t_ms-id]
    return
        count($witness/preceding-sibling::fn:string) + 1
    
};

(:~ 
 : Each token (= relevant word) is encoded in a tei:w which has an @xml:id
 : attribute.
 : In some cases, we don't have any tokens on a page, e.g. when a page only has
 : colophone text.
 : 
 : @param $teixml-uri The base URI of the document, e.g. "12345"
 : @param $page The current page as provided in tei:pb/@n
 :)
declare function vars:get-token-ids-on-page($teixml-uri as xs:string,
    $page as xs:string)
as xs:string* {
    let $page-chunks := commons:get-transcription-and-transliteration-per-page($teixml-uri, $page)
    return
        if (count($page-chunks) gt 1) then
            $page-chunks[descendant::tei:text/@xml:lang = "ara"]//tei:w/@xml:id
        else
            $page-chunks//tei:w/@xml:id
};

declare function vars:get-files-relevant-for-page($relevant-files-for-ms-id as item()+,
    $tokens as xs:string*)
as item()* {
    if (count($tokens) gt 0) then
        let $first-token := $tokens[1]
        let $last-token := $tokens[last()]
        for $file in $relevant-files-for-ms-id return
            let $ser_file := string($file)
            return
                if (contains($ser_file, $first-token) or contains ($ser_file, $last-token)) then
                    $file
                else
                    ()
    else
        util:log-system-out("No tokens found for this page.")
};

(:~
 : Returns a sequence of index positions that are relevant for the current witness
 : and the current page.
 : If a witness doesn't have an entry for the given page in the collation result or if the entry is
 : not on the current page, no index position is returned.
 : 
 : @param $table The CollateX results
 : @param $no-of-sequences The number of result sequences in $table
 : @param $ms-id-position An integer determining which position within the result entry is relevant for the current witness
 : @param $tokens A list of tokens relevant for the current page
 : 
 : @return The sequences in the table where an entry for the current manuscript
 :  and the its current page is available. Zero if there are no entries in the current
 :  CollateX result for the current page.
 :)
declare function vars:get-indices-relevant-for-page($file as document-node(),
    $ms-id-position as xs:integer,
    $tokens as xs:string+)
as xs:integer* {
    let $table := local:get-table($file)
    let $seqs-with-ms-id-entry := $table/fn:array[fn:array[$ms-id-position][fn:map]]
    
    for $e in $seqs-with-ms-id-entry return
        (: get all availabe token IDs in the given sequence for the current manuscript.
        if there is a relevant token, return the sequence index (= $iii) :)
    
        let $all-tokens-in-sequence := $e/fn:array[$ms-id-position]/fn:map/fn:string[@key = 'id']/text()
        return if ($all-tokens-in-sequence = $tokens) then
            index-of($seqs-with-ms-id-entry, $e)
        else
            ()
};

declare function vars:get-non-ms-id-positions-in-array($file as item(),
    $witness-position as xs:integer)
as xs:integer+ {
    let $no-of-witnesses := count($file/fn:map/fn:array[@key = 'witnesses']/fn:string)
    let $no-of-witnesses-as-sequence :=
        for $iii in 1 to $no-of-witnesses return
            $iii
    return
        distinct-values($no-of-witnesses-as-sequence[not(. = $witness-position)]) (: eq functx:value-except#2 :)
};

(:~
 : Creates a map in for every token on a page in which all variants are listed.
 :
 : @param $file The CollateX results
 : @param $entry-pos The current index
 : @param $no-of-sequences The number of result sequences in $table
 : @param $ms-id-position An integer determining which position within the result entry is relevant for the current witness
 : @param $tokens A list of tokens relevant for the current page
 :)
declare function vars:make-map-for-token($file as item(),
    $entry-pos as xs:integer,
    $ms-id-position as xs:integer,
    $non-ms-id-positions as xs:integer*)
as map(*)* {
    let $table := local:get-table($file)
    let $sequence-entry := $table/fn:array[$entry-pos]
    let $ms-id-entry := $sequence-entry/fn:array[$ms-id-position]
    for $token-no in 1 to count($ms-id-entry/fn:map) return
        map {
            "current": map {
                            "t": $ms-id-entry/fn:map[$token-no]/fn:string[@key = 't']/string(),
                            "id": $ms-id-entry/fn:map[$token-no]/fn:string[@key = 'id']/string()
                        },
            "variants":
                for $jjj in $non-ms-id-positions return
                    map {
                        "witness": $file/fn:map/fn:array[@key = 'witnesses']/fn:string[$jjj]/string(),
                        "entry": 
                            if ($sequence-entry/fn:array[$jjj]/fn:map) then
                                $sequence-entry/fn:array[$jjj]/fn:map/fn:string[@key = 't']/string()
                            else
                                []
                    }
        }
};

declare function vars:get-body-object($single-variant as map(*))
as map() {
    map {
        "type": "TextualBody",
        "value": vars:make-annotation-value($single-variant),
        "format": "text/plain",
        "x-content-type": "Variant"
    }
};

declare function vars:make-annotation-value($single-variant as map(*))
as map(*)+ {
        let $entry := map:get($single-variant, "entry") 
        let $witness := map:get($single-variant, "witness") => replace("t_", "")
        
        return
            map {
                "entry": if ($entry instance of xs:string) then $entry else "omisit",
                "witness": map:get($commons:idno-to-sigils-map, $witness)
            }
};

declare function vars:get-target-information($teixml-uri as xs:string,
    $tokenId as xs:string,
    $page as element(tei:TEI))
as map(*) {
    map {
        "selector": map {
            "type": "CssSelector",
            "value": "#" || $tokenId
        },
        "source": "PLACEHOLDER \u2192 TARGET CONTENT HTML IRI"
    }
};

declare function local:get-table($file as item()) {
    $file/fn:map/fn:array[@key = 'table']
};
